/* -*-c-*-
 *
 * $Id: icrypt.h,v 1.4 1998/01/12 16:46:03 mdw Exp $
 *
 * Higher level encryption functions
 *
 * (c) 1998 Mark Wooding
 */

/*----- Licensing notice --------------------------------------------------*
 *
 * This file is part of `become'
 *
 * `Become' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * `Become' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with `become'; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*----- Revision history --------------------------------------------------*
 *
 * $Log: icrypt.h,v $
 * Revision 1.4  1998/01/12 16:46:03  mdw
 * Fix copyright date.
 *
 * Revision 1.3  1997/09/26 09:14:58  mdw
 * Merged blowfish branch into trunk.
 *
 * Revision 1.2.2.1  1997/09/26 09:08:08  mdw
 * Use the Blowfish encryption algorithm instead of IDEA.  This is partly
 * because I prefer Blowfish (without any particularly strong evidence) but
 * mainly because IDEA is patented and Blowfish isn't.
 *
 * Revision 1.2  1997/08/04 10:24:22  mdw
 * Sources placed under CVS control.
 *
 * Revision 1.1  1997/07/21  13:47:49  mdw
 * Initial revision
 *
 */

#ifndef ICRYPT_H
#define ICRYPT_H

#ifdef __cplusplus
  extern "C" {
#endif

/*----- Required headers --------------------------------------------------*/

#include <stddef.h>

#ifndef BLOWFISH_H
#  include "blowfish.h"
#endif

#ifndef CONFIG_H
#  include "config.h"
#endif

/*----- Type definitions --------------------------------------------------*/

/* --- @icrypt_job@ --- */

typedef struct icrypt_job {
  blowfish_key k;			/* Key for en/decrypting */
  int i;				/* Index into the IV buffer */
  unsigned char iv[BLOWFISH_BLKSIZE];	/* IV bytes for encrypting */
} icrypt_job;

/*----- Functions provided ------------------------------------------------*/

extern void icrypt_init(icrypt_job */*j*/,
			unsigned char */*k*/,
			size_t /*sz*/,
			const unsigned char */*iv*/);

/* --- @icrypt_encrypt@ --- *
 *
 * Arguments:	@icrypt_job *j@ = job handle
 *		@const void *src@ = pointer to source buffer
 *		@void *dest@ = pointer to destination buffer
 *		@size_t sz@ = size of buffers to handle
 *
 * Returns:	---
 *
 * Use:		Encrypts data from the source to the destination, using the
 *		key attached to the job handle.
 */

extern void icrypt_encrypt(icrypt_job */*j*/, const void */*src*/,
			   void */*dest*/, size_t /*sz*/);

/* --- @icrypt_decrypt@ --- *
 *
 * Arguments:	@icrypt_job *j@ = job handle
 *		@const void *src@ = pointer to source buffer
 *		@void *dest@ = pointer to destination buffer
 *		@size_t sz@ = size of buffers to handle
 *
 * Returns:	---
 *
 * Use:		Decrypts data from the source to the destination, using
 *		the key attached to the job handle.
 */

extern void icrypt_decrypt(icrypt_job */*j*/, const void */*src*/,
			   void */*dest*/, size_t /*sz*/);

/* --- @icrypt_reset@ --- *
 *
 * Arguments:	@icrypt_job *j@ = pointer to job context block
 *		@unsigned char *k@ = pointer to key data, or zero for
 *			no change
 *		@size_t sz@ = size of the key in bytes
 *		@const unsigned char *iv@ = pointer to IV, or zero
 *
 * Returns:	---
 *
 * Use:		Alters the context block.  This can be used after recovery
 *		of a session key, for example.
 */

extern void icrypt_reset(icrypt_job */*j*/,
			 unsigned char */*k*/,
			 size_t /*sz*/,
			 const unsigned char */*iv*/);

/* --- @icrypt_saveIV@ --- *
 *
 * Arguments:	@icrypt_job *j@ = pointer to job context block
 *		@unsigned char *iv@ = where to store the IV
 *
 * Returns:	---
 *
 * Use:		Writes out the job's IV after munging it a little.
 */

extern void icrypt_saveIV(icrypt_job */*j*/, unsigned char */*iv*/);

/*----- That's all, folks -------------------------------------------------*/

#ifdef __cplusplus
  }
#endif

#endif
