/* -*-c-*-
 *
 * $Id: sym.c,v 1.4 1998/01/12 16:46:28 mdw Exp $
 *
 * Symbol table management
 *
 * (c) 1998 Straylight
 */

/*----- Licensing notice --------------------------------------------------*
 *
 * This file is part of `become'
 *
 * `Become' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * `Become' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with `become'; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*----- Revision history --------------------------------------------------*
 *
 * $Log: sym.c,v $
 * Revision 1.4  1998/01/12 16:46:28  mdw
 * Fix copyright date.
 *
 * Revision 1.3  1997/08/20  16:22:59  mdw
 * Patch memory leak.
 *
 * Revision 1.2  1997/08/04 10:24:25  mdw
 * Sources placed under CVS control.
 *
 * Revision 1.1  1997/07/21  13:47:44  mdw
 * Initial revision
 *
 */

/*----- Header files ------------------------------------------------------*/

/* --- ANSI headers --- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* --- Local headers --- */

#include "sym.h"
#include "utils.h"

/*----- Tuning parameters -------------------------------------------------*/

/* --- Initial hash table size --- *
 *
 * This is the initial @mask@ value.  It must be of the form %$2^n - 1$%,
 * so that it can be used to mask of the bottom bits of a hash value.
 */

#define sym__iSize 255			/* Size of a new hash table */

/* --- Maximum load factor --- *
 *
 * This parameter controls how much the table has to be loaded before the
 * table is extended.  The number of elements %$n$%, the number of bins %$b$%
 * and the limit %$l$% satisfy the relation %$n < bl$%; if a new item is
 * added to the table and this relation is found to be false, the table is
 * doubled in size.
 *
 * The parameter @sym__limFactor@ is the value of %$l$% multiplied by 256;
 * fixed point arithmetic is used to calculate the allowable load.  Note
 * that the exact calculation of this criterion isn't important; what's more
 * significant is that the parameter allows tuning of the rehashing process.
 *
 * The current value is %$3 \over 4$%, which appears to be reasonable on the
 * face of things.
 */

#define sym__limFactor 0x0C0		/* Load factor for growing table */

/*----- CRC table ---------------------------------------------------------*/

/*****************************************************************/
/*                                                               */
/* CRC LOOKUP TABLE                                              */
/* ================                                              */
/*                                                               */
/* The following CRC lookup table was generated automagically    */
/* by `crcgen', which is based on the Rocksoft^tm Model CRC      */
/* Algorithm Table Generation Program.  The model parameters     */
/* supplied to `crcgen' were:                                    */
/*                                                               */
/*    Width   : 32 bits                                          */
/*    Poly    : 0x04C11DB7                                       */
/*    Init    : 0xFFFFFFFF                                       */
/*    XorOut  : 0xFFFFFFFF                                       */
/*    Reverse : Yes                                              */
/*    Check   : 0xCBF43926                                       */
/*                                                               */
/* For more information on the Rocksoft^tm Model CRC Algorithm,  */
/* see the document titled `A Painless Guide to CRC Error        */
/* Detection Algorithms' by Ross Williams                        */
/* (ross@@guest.adelaide.edu.au.). This document is likely to be */
/* in the FTP archive `ftp.adelaide.edu.au/pub/rocksoft'.        */
/*                                                               */
/*****************************************************************/

static unsigned long  sym__crcTable[256] = {
  0x00000000L, 0x77073096L, 0xEE0E612CL, 0x990951BAL,
  0x076DC419L, 0x706AF48FL, 0xE963A535L, 0x9E6495A3L,
  0x0EDB8832L, 0x79DCB8A4L, 0xE0D5E91EL, 0x97D2D988L,
  0x09B64C2BL, 0x7EB17CBDL, 0xE7B82D07L, 0x90BF1D91L,
  0x1DB71064L, 0x6AB020F2L, 0xF3B97148L, 0x84BE41DEL,
  0x1ADAD47DL, 0x6DDDE4EBL, 0xF4D4B551L, 0x83D385C7L,
  0x136C9856L, 0x646BA8C0L, 0xFD62F97AL, 0x8A65C9ECL,
  0x14015C4FL, 0x63066CD9L, 0xFA0F3D63L, 0x8D080DF5L,
  0x3B6E20C8L, 0x4C69105EL, 0xD56041E4L, 0xA2677172L,
  0x3C03E4D1L, 0x4B04D447L, 0xD20D85FDL, 0xA50AB56BL,
  0x35B5A8FAL, 0x42B2986CL, 0xDBBBC9D6L, 0xACBCF940L,
  0x32D86CE3L, 0x45DF5C75L, 0xDCD60DCFL, 0xABD13D59L,
  0x26D930ACL, 0x51DE003AL, 0xC8D75180L, 0xBFD06116L,
  0x21B4F4B5L, 0x56B3C423L, 0xCFBA9599L, 0xB8BDA50FL,
  0x2802B89EL, 0x5F058808L, 0xC60CD9B2L, 0xB10BE924L,
  0x2F6F7C87L, 0x58684C11L, 0xC1611DABL, 0xB6662D3DL,
  0x76DC4190L, 0x01DB7106L, 0x98D220BCL, 0xEFD5102AL,
  0x71B18589L, 0x06B6B51FL, 0x9FBFE4A5L, 0xE8B8D433L,
  0x7807C9A2L, 0x0F00F934L, 0x9609A88EL, 0xE10E9818L,
  0x7F6A0DBBL, 0x086D3D2DL, 0x91646C97L, 0xE6635C01L,
  0x6B6B51F4L, 0x1C6C6162L, 0x856530D8L, 0xF262004EL,
  0x6C0695EDL, 0x1B01A57BL, 0x8208F4C1L, 0xF50FC457L,
  0x65B0D9C6L, 0x12B7E950L, 0x8BBEB8EAL, 0xFCB9887CL,
  0x62DD1DDFL, 0x15DA2D49L, 0x8CD37CF3L, 0xFBD44C65L,
  0x4DB26158L, 0x3AB551CEL, 0xA3BC0074L, 0xD4BB30E2L,
  0x4ADFA541L, 0x3DD895D7L, 0xA4D1C46DL, 0xD3D6F4FBL,
  0x4369E96AL, 0x346ED9FCL, 0xAD678846L, 0xDA60B8D0L,
  0x44042D73L, 0x33031DE5L, 0xAA0A4C5FL, 0xDD0D7CC9L,
  0x5005713CL, 0x270241AAL, 0xBE0B1010L, 0xC90C2086L,
  0x5768B525L, 0x206F85B3L, 0xB966D409L, 0xCE61E49FL,
  0x5EDEF90EL, 0x29D9C998L, 0xB0D09822L, 0xC7D7A8B4L,
  0x59B33D17L, 0x2EB40D81L, 0xB7BD5C3BL, 0xC0BA6CADL,
  0xEDB88320L, 0x9ABFB3B6L, 0x03B6E20CL, 0x74B1D29AL,
  0xEAD54739L, 0x9DD277AFL, 0x04DB2615L, 0x73DC1683L,
  0xE3630B12L, 0x94643B84L, 0x0D6D6A3EL, 0x7A6A5AA8L,
  0xE40ECF0BL, 0x9309FF9DL, 0x0A00AE27L, 0x7D079EB1L,
  0xF00F9344L, 0x8708A3D2L, 0x1E01F268L, 0x6906C2FEL,
  0xF762575DL, 0x806567CBL, 0x196C3671L, 0x6E6B06E7L,
  0xFED41B76L, 0x89D32BE0L, 0x10DA7A5AL, 0x67DD4ACCL,
  0xF9B9DF6FL, 0x8EBEEFF9L, 0x17B7BE43L, 0x60B08ED5L,
  0xD6D6A3E8L, 0xA1D1937EL, 0x38D8C2C4L, 0x4FDFF252L,
  0xD1BB67F1L, 0xA6BC5767L, 0x3FB506DDL, 0x48B2364BL,
  0xD80D2BDAL, 0xAF0A1B4CL, 0x36034AF6L, 0x41047A60L,
  0xDF60EFC3L, 0xA867DF55L, 0x316E8EEFL, 0x4669BE79L,
  0xCB61B38CL, 0xBC66831AL, 0x256FD2A0L, 0x5268E236L,
  0xCC0C7795L, 0xBB0B4703L, 0x220216B9L, 0x5505262FL,
  0xC5BA3BBEL, 0xB2BD0B28L, 0x2BB45A92L, 0x5CB36A04L,
  0xC2D7FFA7L, 0xB5D0CF31L, 0x2CD99E8BL, 0x5BDEAE1DL,
  0x9B64C2B0L, 0xEC63F226L, 0x756AA39CL, 0x026D930AL,
  0x9C0906A9L, 0xEB0E363FL, 0x72076785L, 0x05005713L,
  0x95BF4A82L, 0xE2B87A14L, 0x7BB12BAEL, 0x0CB61B38L,
  0x92D28E9BL, 0xE5D5BE0DL, 0x7CDCEFB7L, 0x0BDBDF21L,
  0x86D3D2D4L, 0xF1D4E242L, 0x68DDB3F8L, 0x1FDA836EL,
  0x81BE16CDL, 0xF6B9265BL, 0x6FB077E1L, 0x18B74777L,
  0x88085AE6L, 0xFF0F6A70L, 0x66063BCAL, 0x11010B5CL,
  0x8F659EFFL, 0xF862AE69L, 0x616BFFD3L, 0x166CCF45L,
  0xA00AE278L, 0xD70DD2EEL, 0x4E048354L, 0x3903B3C2L,
  0xA7672661L, 0xD06016F7L, 0x4969474DL, 0x3E6E77DBL,
  0xAED16A4AL, 0xD9D65ADCL, 0x40DF0B66L, 0x37D83BF0L,
  0xA9BCAE53L, 0xDEBB9EC5L, 0x47B2CF7FL, 0x30B5FFE9L,
  0xBDBDF21CL, 0xCABAC28AL, 0x53B39330L, 0x24B4A3A6L,
  0xBAD03605L, 0xCDD70693L, 0x54DE5729L, 0x23D967BFL,
  0xB3667A2EL, 0xC4614AB8L, 0x5D681B02L, 0x2A6F2B94L,
  0xB40BBE37L, 0xC30C8EA1L, 0x5A05DF1BL, 0x2D02EF8DL,
};

/*****************************************************************/
/*                   End of CRC Lookup Table                     */
/*****************************************************************/

/* --- Work out the CRC of a bytestring --- */

#define sym__crc(s_, l_, hv_) do {					\
  unsigned long h_ = 0xFFFFFFFFL;					\
  const char *p_ = s_;							\
  const char *e_ = (s_) + (l_);						\
									\
  while (p_ < e_)							\
    h_ = (h_ >> 8) ^ (sym__crcTable[(*p_++ ^ h_) & 0xFF]);		\
  hv_ = ~h_ & 0xFFFFFFFFL;						\
} while (0)

/*----- Main code ---------------------------------------------------------*/

#ifndef DEBUG_CRC

/* --- @sym_createTable@ --- *
 *
 * Arguments:	@sym_table *t@ = symbol table to initialise
 *
 * Returns:	---
 *
 * Use:		Initialises the given symbol table.
 */

void sym_createTable(sym_table *t)
{
  size_t i;

  t->mask = sym__iSize;
  t->c = (sym__iSize * sym__limFactor) >> 8;
  t->a = xmalloc((t->mask + 1) * sizeof(sym_base *));

  for (i = 0; i < sym__iSize + 1; i++)
    t->a[i] = 0;  
}

/* --- @sym_destroyTable@ --- *
 *
 * Arguments:	@sym_table *t@ = pointer to symbol table in question
 *
 * Returns:	---
 *
 * Use:		Destroys a symbol table, freeing all the memory it used to
 *		occupy.
 */

void sym_destroyTable(sym_table *t)
{
  size_t i;
  sym_base *p, *q;

  for (i = 0; i <= t->mask; i++) {
    p = t->a[i];
    while (p) {
      q = p->next;
      free(p->name);
      free(p);
      p = q;
    }
  }
  free(t->a);
}

/* --- @sym_find@ --- *
 *
 * Arguments:	@sym_table *t@ = pointer to symbol table in question
 *		@const char *n@ = pointer to symbol table to look up
 *		@long l@ = length of the name string or negative to measure
 *		@size_t sz@ = size of desired symbol object, or zero
 *		@unsigned *f@ = pointer to a flag, or null.
 *
 * Returns:	The address of a @sym_base@ structure, or null if not found
 *		and @sz@ is zero.
 *
 * Use:		Looks up a symbol in a given symbol table.  The name is
 *		passed by the address of its first character.  The length
 *		may be given, in which case the name may contain arbitrary
 *		binary data, or it may be given as a negative number, in
 *		which case the length of the name is calculated as
 *		@strlen(n)@.
 *
 *		The return value is the address of a pointer to a @sym_base@
 *		block (which may have other things on the end, as above).  If
 *		the symbol could be found, the return value points to the
 *		symbol block.  If the symbol wasn't there, then if @sz@ is
 *		nonzero, a new symbol is created and its address is returned;
 *		otherwise a null pointer is returned.
 *
 *		The value of @*f@ indicates whether a new symbol entry was
 *		created: a nonzero value indicates that an old value was
 *		found.
 */

void *sym_find(sym_table *t, const char *n, long l, size_t sz, unsigned *f)
{
  unsigned long hash;			/* Hash value for user's name */
  size_t len = l < 0 ? strlen(n) + 1 : l; /* Find length of user's name */
  sym_base *bin;			/* Bin containing our item */
  sym_base *p, *q;			/* Pointer wandering through list */

  /* --- Find the correct bin --- */

  sym__crc(n, len, hash);		/* Find hash value for this name */
  bin = p = (sym_base *)(t->a + (hash & t->mask));

  /* --- Search the bin list --- */

  while (p->next) {
    if (hash == p->next->hash &&	/* Check hash values first */
	len == p->next->len &&		/* Then compare string lengths */
	!memcmp(n, p->next->name, len))	/* And finally compare the strings */
    {
      /* --- Found a match --- *
       *
       * As a minor, and probably pointless, tweak, move the item to the
       * front of its bin list.
       */

      q = p->next;			/* Find the actual symbol block */
      p->next = q->next;		/* Extract block from bin list */
      q->next = bin->next;		/* Set up symbol's next pointer */
      bin->next = q;			/* And reinsert the block */

      /* --- Return the block --- */

      if (f) *f = 1;			/* Didn't fail to find the item */
      return (q);			/* And return the block */
    }

    p = p->next;			/* Move onto the next item */
  }

  /* --- Couldn't find the item there --- */

  if (f) *f = 0;			/* Failed to find the block */
  if (!sz) return (0);			/* Return zero if not creating */

  /* --- Create a new symbol block and initialise it --- */

  p = xmalloc(sz);			/* Create a new symbol block */
  p->next = bin->next;			/* Set up the next pointer */
  p->hash = hash;			/* Set up the hash value too */
  p->name = xmalloc(len);		/* Allocate a block for the name */
  p->len = len;				/* And set up the string length */
  memcpy(p->name, n, len);		/* And copy the string over */

  bin->next = p;			/* Put the pointer into the bin */

  /* --- Consider growing the array --- */

  if (!--t->c) {
    unsigned long m = t->mask + 1;	/* Next set bit in has word */
    sym_base *p, *q, *r;		/* More useful pointers */
    size_t i, lim;			/* Loop counter and limit */

    /* --- Update values in the anchor block --- */

    t->c = ((t->mask + 1) * sym__limFactor) >> 8; /* Set load value */
    t->mask = (t->mask + 1) * 2 - 1;	/* Set the new mask value */
    t->a = xrealloc(t->a, (t->mask + 1) * sizeof(sym_base *));

    /* --- Now wander through the table rehashing things --- *
     *
     * This loop is very careful to avoid problems with aliasing.  The items
     * are dealt with from the end backwards to avoid overwriting bins before
     * they've been processed.
     */

    lim = (t->mask + 1) >> 1;
    for (i = 0; i < lim; i++) {
      /* --- Some initialisation --- */

      r = t->a[i];			/* Find the list we're dissecting */
      p = (sym_base *)(t->a + i);	/* Find bit-clear list */
      q = (sym_base *)(t->a + i + lim);	/* And the bit-set lsit */

      /* --- Now go through the @r@ list --- */

      while (r) {
	if (r->hash & m)		/* Is the next bit set? */
	  q = q->next = r;		/* Yes, so fit up previous link */
	else
	  p = p->next = r;		/* No, so fit up previous link */
	r = r->next;			/* Move onto the next item */
      }
      p->next = q->next = 0;		/* Null terminate the new lists */
    }
  }

  /* --- Finished that, so return the new symbol block --- */

  return (p);
}

/* --- @sym_remove@ --- *
 *
 * Arguments:	@sym_table *i@ = pointer to a symbol table object
 *		@void *b@ = pointer to symbol table entry
 *
 * Returns:	---
 *
 * Use:		Removes the object from the symbol table.  The space occupied
 *		by the object and its name is freed; anything else attached
 *		to the entry should already be gone by this point.
 */

void sym_remove(sym_table *t, void *b)
{
  /* --- A quick comment --- *
   *
   * Since the @sym_base@ block contains the hash, finding the element in the
   * bin list is really quick -- it's not worth bothering with things like
   * doubly linked lists.
   */

  sym_base *p = b;
  sym_base *bin = (sym_base *)(t->a + (p->hash & t->mask));

  /* --- Find the item in the bin list --- */

  while (bin->next) {
    if (bin->next == p)
      break;
    bin = bin->next;
  }
  if (!bin->next)
    return;

  /* --- Now just remove the item from the list and free it --- *
   *
   * Oh, and bump the load counter.
   */

  bin->next = p->next;
  free(p->name);
  free(p);
  t->c++;
}

/* --- @sym_createIter@ --- *
 *
 * Arguments:	@sym_iter *i@ = pointer to an iterator object
 *		@sym_table *t@ = pointer to a symbol table object
 *
 * Returns:	---
 *
 * Use:		Creates a new symbol table iterator which may be used to
 *		iterate through a symbol table.
 */

void sym_createIter(sym_iter *i, sym_table *t)
{
  i->t = t;
  i->i = 0;
  i->n = 0;
}

/* --- @sym_next@ --- *
 *
 * Arguments:	@sym_iter *i@ = pointer to iterator object
 *
 * Returns:	Pointer to the next symbol found, or null when finished.
 *
 * Use:		Returns the next symbol from the table.  Symbols are not
 *		returned in any particular order.
 */

void *sym_next(sym_iter *i)
{
  sym_base *p;

  /* --- Find the next item --- */

  while (!i->n) {
    if (i->i > i->t->mask)
      return (0);
    i->n = i->t->a[i->i++];
  }

  /* --- Update the iterator block --- */

  p = i->n;
  i->n = p->next;

  /* --- Done --- */

  return (p);
}

#endif

/*----- CRC test code -----------------------------------------------------*/

#ifdef DEBUG_CRC

int main(void)
{
  unsigned long h;
  sym__crc("123456789", 9, h);
  printf("crc check == %04x\n", h);
  return (0);
}

#endif

/*----- Symbol table test code --------------------------------------------*/

#ifdef TEST_RIG

#include <errno.h>
#include <time.h>

#include "dbutils.h"

typedef struct sym_word {
  sym_base base;
  size_t i;
} sym_word;


/* --- What it does --- *
 *
 * Reads the file /usr/dict/words (change to some other file full of
 * interesting and appropriate bits of text to taste) into a big buffer and
 * picks apart into lines.  Then picks lines at random and enters them into
 * the symbol table.
 */

int main(void)
{
  char *buff, *p, *lim;
  size_t sz, done;
  FILE *fp;
  int i;
  char **line;
  sym_word **flag;
  sym_table tbl;
  int entries;

  /* --- Initialise for reading the file --- */

  sz = BUFSIZ;
  buff = xmalloc(sz + 1);
  done = 0;

  if ((fp = fopen("/usr/dict/words", "r")) == 0)
    fprintf(stderr, "buggered ;-( (%s)\n", strerror(errno));

  /* --- Read buffers of text --- *
   *
   * Read a buffer.  If more to come, double the buffer size and try again.
   * This is the method I recommended to comp.lang.c, so I may as well try
   * it.
   */

  for (;;) {
    i = fread(buff + done, 1, sz - done, fp);
    done += i;
    if (done != sz)
      break;
    sz <<= 1;
    buff = xrealloc(buff, sz + 1);
  }

  /* --- Count the lines --- */

  lim = buff + done;

  sz = 1;
  for (p = buff; p < lim; p++)
    if (*p == '\n') sz++;

  /* --- Build a table of line starts --- */

  line = xmalloc(sz * sizeof(char *));
  i = 0;
  line[i++] = buff;
  for (p = buff; p < lim; p++)
    if (*p == '\n') *p = 0, line[i++] = p + 1;
  *lim = 0;

  /* --- Build a table of lines --- *
   *
   * This reverses the mapping which the symbol table performs, so that its
   * accuracy can be tested.
   */

  flag = xmalloc(sz * sizeof(sym_word *));
  for (i = 0; i < sz; i++)
    flag[i] = 0;
  entries = 0;

  sym_createTable(&tbl);

  for (;;) {
    i = (unsigned)rand() % sz;

    switch (rand() % 5)
    {
      case 0: {
	sym_word *w;

	/* printf("find `%s'\n", line[i]); */
	if ((rand() & 1023) == 0) {
	  putchar('.'); fflush(stdout);
	}

	w = sym_find(&tbl, line[i], -1, 0, 0);
	if (w != flag[i])
	  printf("*** error: find `%s' gave %p not %p\n",
		 line[i], (void *)w, (void *)flag[i]);
	else if (w && w->i != i)
	  printf("*** error: find sym for `%s' gives index %i not %i\n",
		 line[i], w->i, i);	
      } break;

      case 1: {
	unsigned f;
	sym_word *w;

	/* printf("create `%s'\n", line[i]); */
	if ((rand() & 1023) == 0) {
	  putchar('+'); fflush(stdout);
	}

	w = sym_find(&tbl, line[i], -1, sizeof(sym_word), &f);
	if (f)
	{
	  if (w != flag[i])
	    printf("*** error: create `%s' gave %p not %p\n",
		   line[i], (void *)w, (void *)flag[i]);
	  else if (w && w->i != i)
	    printf("*** error: create sym for `%s' gives index %i not %i\n",
		   line[i], w->i, i);
	}
	else
	{
	  if (flag[i])
	    printf("*** error: create `%s' gave new block, should be %p\n",
		   line[i], (void *)flag[i]);
	  else {
	    flag[i] = w;
	    w->i = i;
	    entries++;
	  }
	}
      } break;

      case 2: {
	sym_iter it;
	sym_word *w, **ntbl;
	int v = (rand() % entries) == 0;
	if (!v)
	  break;
	printf("\niterated %i entries\n", entries);
	break;

	printf("iterate\n");

	ntbl = xmalloc(sz * sizeof(sym_word *));
	memcpy(ntbl, flag, sz * sizeof(sym_word *));
	sym_createIter(&it, &tbl);

	while ((w = sym_next(&it)) != 0) {
	  if (ntbl[w->i] == 0)
	    printf("*** error: iterate returned duff item %i\n", w->i);
	  else
	    ntbl[w->i] = 0;
	}

	for (i = 0; i < sz; i++)
	  if (ntbl[i]) printf("*** error: iterate didn't return item %i\n",
			      i);
	free(ntbl);
      } break;

      case 3: {
	sym_base *b;
	int v = rand() & 255 ? 0 : 1;
	break;

	printf("dump\n");

	for (i = 0; i <= tbl.mask; i++) {
	  if (!tbl.a[i]) continue;
	  if (v) printf("  %i: ", i);
	  b = tbl.a[i];
	  while (b) {
	    if ((b->hash & tbl.mask) != i)
	      printf("*** error: bad hash value found");
	    if (v) printf("`%s'(%08lx:%lu) ",
			  line[((sym_word *)b)->i],
			  b->hash,
			  b->hash & tbl.mask);
	    b = b->next;
	  }
	  if (v) putchar('\n');
	}
      } break;

      case 4: {
	if (flag[i]) {
	  /* printf("remove `%s'\n", flag[i]->base.name); */
	  if ((rand() & 1023) == 0) {
	    putchar('-'); fflush(stdout);
	  }
	  sym_remove(&tbl, flag[i]);
	  flag[i] = 0;
	  entries--;
	}
      } break;
    }

  }

  return (0);
}

#endif

/*----- That's all, folks -------------------------------------------------*/
